# Generated by Django 4.2 on 2023-11-22 03:45

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("flex_report", "0001_initial"),
    ]

    operations = [
        migrations.AlterField(
            model_name="column",
            name="title",
            field=models.CharField(db_index=True, max_length=150, verbose_name="Title"),
        ),
    ]
